# type: ignore
__submodules__ = ["hook", "run", "args"]

# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = ["ArgumentGroups", "ArgumentParserArgs", "SnakebidsConfig", "app", "hookimpl"]
# </AUTOGEN_INIT>
