"""
This type stub file was generated by pyright.
"""

if False: ...

class Version:
    def __init__(self, major: int, minor: int) -> None: ...
    @property
    def major(self) -> int: ...
    @property
    def minor(self) -> int: ...
    def __eq__(self, v: Any) -> bool: ...
    def __lt__(self, v: Version) -> bool: ...
    def __le__(self, v: Version) -> bool: ...
    def __gt__(self, v: Version) -> bool: ...
    def __ge__(self, v: Version) -> bool: ...

def version(
    major: int | str | Tuple[int, int] | None, minor: Optional[int] = ...
) -> Optional[Version]: ...

class Tag:
    def __init__(self, handle: str, prefix: str) -> None: ...
    @property
    def handle(self) -> str: ...
    @property
    def prefix(self) -> str: ...

class DocInfo:
    """
    Store document information, can be used for analysis of a loaded YAML document
    requested_version: if explicitly set before load
    doc_version: from %YAML directive
    tags: from %TAG directives in scanned order
    """
    def __init__(
        self,
        requested_version: Optional[Version] = ...,
        doc_version: Optional[Version] = ...,
        tags: Optional[list[Tag]] = ...,
    ) -> None: ...
