# <AUTOGEN_INIT>
from ._config import (
    set_bids_spec,
)
from ._factory import (
    BidsFunction,
    bids_factory,
)
from ._presets import (
    bids,
)
from ._utils import (
    BidsPathEntitySpec,
    BidsPathSpec,
    BidsPathSpecFile,
)

__all__ = [
    "BidsFunction",
    "BidsPathEntitySpec",
    "BidsPathSpec",
    "BidsPathSpecFile",
    "bids",
    "bids_factory",
    "set_bids_spec",
]
# </AUTOGEN_INIT>
