"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import YAMLError

if False: ...
__all__ = ["Serializer", "SerializerError"]

class SerializerError(YAMLError): ...

class Serializer:
    ANCHOR_TEMPLATE = ...
    ANCHOR_RE = ...
    def __init__(
        self,
        encoding: Any = ...,
        explicit_start: Optional[bool] = ...,
        explicit_end: Optional[bool] = ...,
        version: Optional[VersionType] = ...,
        tags: Any = ...,
        dumper: Any = ...,
    ) -> None: ...
    @property
    def emitter(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    def open(self) -> None: ...
    def close(self) -> None: ...
    def serialize(self, node: Any) -> None: ...
    def anchor_node(self, node: Any) -> None: ...
    def generate_anchor(self, node: Any) -> Any: ...
    def serialize_node(self, node: Any, parent: Any, index: Any) -> None: ...

def templated_id(s: Text) -> Any: ...
