#snakebids\.paths\.utils\.BidsPathSpec,
#snakebids\.types\.InputsConfig,
#snakebids\.types\.ZipList,
#snakebids\.types\.ZipListLike {
  visibility: hidden;
  height: 0;
}

#snakebids\.bidsapp\.run\._Runner ~ dd .field-list {
  display: none
}
