# type: ignore
__submodules__ = ["_factory", "_presets", "_config", "_utils"]

# <AUTOGEN_INIT>
import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

__all__ = [
    "BidsFunction",
    "BidsPathEntitySpec",
    "BidsPathSpec",
    "BidsPathSpecFile",
    "bids",
    "bids_factory",
    "set_bids_spec",
]
# </AUTOGEN_INIT>
