"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = ["ScalarFloat", "ExponentialFloat", "ExponentialCapsFloat"]

class ScalarFloat(float):
    def __new__(cls: Any, *args: Any, **kw: Any) -> Any: ...
    def __iadd__(self, a: Any) -> Any: ...
    def __ifloordiv__(self, a: Any) -> Any: ...
    def __imul__(self, a: Any) -> Any: ...
    def __ipow__(self, a: Any) -> Any: ...
    def __isub__(self, a: Any) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...
    def dump(self, out: Any = ...) -> None: ...

class ExponentialFloat(ScalarFloat):
    def __new__(cls, value: Any, width: Any = ..., underscore: Any = ...) -> Any: ...

class ExponentialCapsFloat(ScalarFloat):
    def __new__(cls, value: Any, width: Any = ..., underscore: Any = ...) -> Any: ...
