"""
This type stub file was generated by pyright.
"""

import datetime

if False: ...

class TimeStamp(datetime.datetime):
    def __init__(self, *args: Any, **kw: Any) -> None: ...
    def __new__(cls, *args: Any, **kw: Any) -> Any: ...
    def __deepcopy__(self, memo: Any) -> Any: ...
    def replace(
        self,
        year: Any = ...,
        month: Any = ...,
        day: Any = ...,
        hour: Any = ...,
        minute: Any = ...,
        second: Any = ...,
        microsecond: Any = ...,
        tzinfo: Any = ...,
        fold: Any = ...,
    ) -> Any: ...
