"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.reader import Reader
from ruamel.yaml.scanner import RoundTripScanner, Scanner
from ruamel.yaml.parser import Parser, RoundTripParser
from ruamel.yaml.composer import Composer
from ruamel.yaml.constructor import (
    BaseConstructor,
    Constructor,
    RoundTripConstructor,
    SafeConstructor,
)
from ruamel.yaml.resolver import VersionedResolver

if False: ...
__all__ = ["BaseLoader", "SafeLoader", "Loader", "RoundTripLoader"]

class BaseLoader(Reader, Scanner, Parser, Composer, BaseConstructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class SafeLoader(Reader, Scanner, Parser, Composer, SafeConstructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class Loader(Reader, Scanner, Parser, Composer, Constructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class RoundTripLoader(
    Reader,
    RoundTripScanner,
    RoundTripParser,
    Composer,
    RoundTripConstructor,
    VersionedResolver,
):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...
