"""
This type stub file was generated by pyright.
"""

if False: ...
anchor_attrib = ...

class Anchor:
    __slots__ = ...
    attrib = ...
    def __init__(self) -> None: ...
    def __repr__(self) -> Any: ...
