"""
This type stub file was generated by pyright.
"""

from __future__ import annotations

from typing import Any

from ruamel.yaml.main import *

from .cyaml import *

if False: ...
_package_data: dict[Any, Any] = ...
version_info = ...
__version__ = ...
__with_libyaml__ = ...
