"""
This type stub file was generated by pyright.
"""

if False: ...
__all__ = ["ScalarBoolean"]

class ScalarBoolean(int):
    def __new__(cls: Any, *args: Any, **kw: Any) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self, any: bool = ...) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...
