"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import YAMLError

if False: ...
__all__ = ["Reader", "ReaderError"]

class ReaderError(YAMLError):
    def __init__(
        self, name: Any, position: Any, character: Any, encoding: Any, reason: Any
    ) -> None: ...
    def __str__(self) -> Any: ...

class Reader:
    def __init__(self, stream: Any, loader: Any = ...) -> None: ...
    def reset_reader(self) -> None: ...
    @property
    def stream(self) -> Any: ...
    @stream.setter
    def stream(self, val: Any) -> None: ...
    def peek(self, index: int = ...) -> Text: ...
    def prefix(self, length: int = ...) -> Any: ...
    def forward_1_1(self, length: int = ...) -> None: ...
    def forward(self, length: int = ...) -> None: ...
    def get_mark(self) -> Any: ...
    def determine_encoding(self) -> None: ...

    NON_PRINTABLE = ...
    _printable_ascii = ...
    def check_printable(self, data: Any) -> None: ...
    def update(self, length: int) -> None: ...
    def update_raw(self, size: Optional[int] = ...) -> None: ...
