"""
This type stub file was generated by pyright.
"""

if False: ...
tag_attrib = ...

class Tag:
    """store original tag information for roundtripping"""

    attrib = ...
    def __init__(
        self, handle: Any = ..., suffix: Any = ..., handles: Any = ...
    ) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: Any) -> bool: ...
    def startswith(self, x: str) -> bool: ...
    @property
    def trval(self) -> Optional[str]: ...

    value = ...
    @property
    def uri_decoded_suffix(self) -> Optional[str]: ...
    def select_transform(self, val: bool) -> None:
        """
        val: False -> non-round-trip
             True -> round-trip
        """
        ...

    def check_handle(self) -> bool: ...
